import unittest

class TopLevelParseStateManagement(unittest.TestCase):
	@classmethod
	def setUpClass(cls):
		cls._harena_t = gdb.lookup_type("HArena")
		cls._arena = gdb.parse_and_eval("malloc(sizeof(HArena))").cast(cls._harena_t.pointer()).dereference() # Quick and dirty way to allocate something in the process GDB is debugging (the inferior)
		cls._hps_t = gdb.lookup_type("HParseState")
		cls._test_parse_state = gdb.parse_and_eval("malloc(sizeof(HParseState))").cast(cls._hps_t.pointer()).dereference()
		# TODO: make this independent of endianness and pointer size
		gdb.selected_inferior().write_memory(int(cls._test_parse_state.address)+(cls._hps_t['arena'].bitpos//8), int(cls._arena.address).to_bytes(8, 'little')) # horrible, horrible way to set the 
		cls._hps_t = cls._test_parse_state.address.dereference() # "refresh" the gdb.Value, since the effects of .write_memory() don't seem to show up

	def setUp(self):
		self.top_level_parse = TopLevelParse()
		self.ps1 = ParserStack(16, self._arena, self.top_level_parse)
		self.parser1 = Parser("foo", 32)

	def test_enter_h_packrat_parse(self):
		self.top_level_parse.enter_h_packrat_parse(32)
		self.assertIsInstance(self.top_level_parse.parser_stacks[0], ParserStack)

	def test_enter_h_do_parse_new_parser(self):
		ps1 = ParserStack(16, self._arena, self.top_level_parse)
		self.top_level_parse.enter_h_packrat_parse(32)
		#simulating effects of enter_h_packrat_parse
		self.top_level_parse.enter_h_do_parse(16, self._arena, 32)
		parser1 = self.top_level_parse.parser_objs[32]
		self.assertIs(self.top_level_parse.parser_stacks[-1].p_stack[-1], self.top_level_parse.parser_objs[32])
		self.assertIsNone(parser1.name)
		self.assertEqual(parser1.address, 32)

	def test_enter_h_do_parse_known_parser(self):
		self.top_level_parse.enter_h_packrat_parse(32)
		parser1 = Parser("foo", 32)
		self.top_level_parse.parser_objs[32] = parser1 # Add the parser into the db by hand
		self.top_level_parse.enter_h_do_parse(16, self._arena, 32)
		self.assertIs(self.top_level_parse.parser_stacks[-1].p_stack[-1], parser1)
		self.assertIs(self.top_level_parse.parser_stacks[-1].p_stack[-1], self.top_level_parse.parser_objs[32])
		self.assertEqual(self.top_level_parse.parser_stacks[-1].p_stack[-1].name, "foo")
		self.assertEqual(self.top_level_parse.parser_stacks[-1].p_stack[-1].address, 32)

	def test_ehd_calls_first_h_do_parse_after_packrat_parse(self):
		parser1 = Parser("foo", 32)
		self.top_level_parse.enter_h_packrat_parse(parser1.address)
		self.top_level_parse.enter_h_do_parse(16, self._arena, parser1.address)
		ps1 = self.top_level_parse.peek_parserstack()
		self.assertEqual(ps1.parse_state, 16)

	def test_first_h_do_parse_after_packrat_parse(self):
		ps1 = ParserStack(None, self._arena, self.top_level_parse)
		self.top_level_parse.parser_stacks.append(ps1)
		self.top_level_parse.first_h_do_parse_after_packrat_parse(16, self._arena)
		self.assertEqual(ps1.parse_state, 16)

	def test_return_from_h_packrat_parse(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.return_from_h_packrat_parse()
		self.assertEqual(self.top_level_parse.parser_stacks, [])

	def test_enter_perform_lowlevel_parse_known_parser(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.top_level_parse.parser_stacks[-1].p_stack.append(self.parser1)
		self.top_level_parse.enter_perform_lowlevel_parse(32)
		self.assertIs(self.top_level_parse.parser_stacks[-1].p_stack[-1], self.parser1)

	def test_enter_perform_lowlevel_parse_new_parser(self):
		# Note: since parsers are pushed on the stack in h_do_parse now, it's virtually guaranteed
		# that their type will be resolved before perform_lowlevel_parse sees them
		# (thus parser_objs will contain the parser, and the branch being tested won't run)
		self.top_level_parse.parser_stacks.append(self.ps1)
		parser1 = self.top_level_parse.enter_perform_lowlevel_parse(32)
		self.assertIsNone(parser1.name)
		self.assertEqual(parser1.address, 32)

	@unittest.skip("Function is a placeholder at the moment")
	def test_return_from_lowlevel_parse(self):
		pass

	# TODO: This is closer to an integration test: when enter_h_arena_malloc_raw is called, the allocation
	# is added to the parser on top of the stack of stacks.
	# Testing here that after calling enter_h_do_parse, then enter_h_arena_malloc_raw
	# the latter will attribute the allocation to the parser we just set
	def test_enter_h_arena_malloc_raw_top_of_stack(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.top_level_parse.enter_h_do_parse(16, self._arena, self.parser1.address)
		self.top_level_parse.enter_h_arena_malloc_raw(256)
		self.assertEqual(self.parser1.bytes_used[int(self._arena.address)], 256)

	@skipIf(self.top_level_parse.memory_stat_method == HammerMemoryStatisticsMethod.DETAILED_ARENA_STATS)
	def test_enter_h_arena_malloc_raw(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1, self._arena)
		self.top_level_parse.enter_h_arena_malloc_raw(256)
		self.assertEqual(self.top_level_parse.parser_stacks[-1].p_stack[-1].bytes_used[int(self._arena.address)], 256)

	def test_enter_h_arena_malloc_raw_no_parser(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.enter_h_arena_malloc_raw(256)
		self.assertEqual(self.top_level_parse.parser_stacks[-1].unclaimed_mem_use, 256)

	def test_enter_h_arena_malloc_raw_no_stack(self):
		self.top_level_parse.enter_h_arena_malloc_raw(256)
		self.assertEqual(self.top_level_parse.unclaimed_mem_use, 256)

	def test_parse_virtual(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		parser2 = Parser(None, 32)
		self.top_level_parse.parser_stacks[-1].push(parser2, self._arena)
		self.top_level_parse.parse_virtual("foo")
		self.assertEqual(parser2.name, "foo")

	def test_peek_parserstack(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		ps1 = self.top_level_parse.peek_parserstack()
		self.assertIs(ps1, self.ps1)

	def test_peek_parser(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1, self._arena)
		parser1 = self.top_level_parse.peek_parser()
		self.assertIs(parser1, self.parser1)

	def test_parser_by_address(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1, self._arena)
		parser1 = self.top_level_parse.parser_by_address(32)

	def test_parsers_by_name(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1, self._arena)
		parser2 = Parser("foo", 48)
		self.top_level_parse.parser_objs[48] = parser2
		self.ps1.push(parser2, self._arena)
		parserlist = self.top_level_parse.parsers_by_name("foo")
		self.assertEqual(parserlist, [self.parser1, parser2])

	def test_add_or_get_parser_add(self):
		parser2 = self.top_level_parse.add_or_get_parser(32)
		self.assertIsNone(parser2.name)
		self.assertEqual(parser2.address, 32)
		self.assertIs(self.top_level_parse.parser_objs[32], parser2)

	def test_add_or_get_parser_get(self):
		self.top_level_parse.parser_objs[32] = self.parser1
		parser1 = self.top_level_parse.add_or_get_parser(32)
		self.assertIs(parser1, self.parser1)

	def test_get_parser_top_per_arena_mem(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1, self._arena)
		parser2 = Parser("bar", 48)
		self.top_level_parse.parser_objs[48] = parser2
		self.ps1.push(parser2, self._arena)
		self.parser1.add_mem_use(16, 50)
		parser2.add_mem_use(16, 60)
		self.assertIs(self.top_level_parse.get_parser_top_per_arena_mem(), parser2)

	def test_get_parser_top_total_arena_mem(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1, self._arena)
		parser2 = Parser("bar", 48)
		self.top_level_parse.parser_objs[48] = parser2
		self.ps1.push(parser2, self._arena)
		self.parser1.add_mem_use(16, 50)
		self.parser1.add_mem_use(128, 50)
		parser2.add_mem_use(16, 60)
		self.assertIs(self.top_level_parse.get_parser_top_total_arena_mem(), self.parser1)

	def test_get_avg_mem_use_per_arena(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1, self._arena)
		parser2 = Parser("bar", 48)
		self.top_level_parse.parser_objs[48] = parser2
		self.ps1.push(parser2, self._arena)
		self.parser1.add_mem_use(16, 50)
		self.parser1.add_mem_use(128, 50)
		parser2.add_mem_use(16, 60)
		self.assertEqual(self.top_level_parse.get_avg_mem_use_per_arena(), {16: 55, 128: 50})

	def test_dump_memory_stats(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1, self._arena)
		parser2 = Parser("bar", 48)
		self.top_level_parse.parser_objs[48] = parser2
		self.ps1.push(parser2, self._arena)
		self.parser1.add_mem_use(16, 50)
		self.parser1.add_mem_use(128, 50)
		parser2.add_mem_use(16, 60)

		stats = self.top_level_parse.dump_memory_stats()
		self.assertEqual(stats, [('foo', '0x20', {16: 50, 128: 50}), ('bar', '0x30', {16: 60})])
